@@ -22,7 +22,7 @@
     You should also have recieved a copy of the Artistic license with
     this program.
 
-   $Id: bdengine.c,v 1.14 2001/10/21 07:03:49 andy Exp $
+   $Id: bdengine.c,v 1.4 2006/01/10 17:27:53 andymort Exp $
    
    We're getting there. At the moment, the file used by default is ~/.birthdays
    under UNIX, or C:\PERSONAL\BDAYS.LST under DOS, but this can be overridden on
@@ -50,6 +50,9 @@
      it's behaviour. In most cases, it will just quietly ignore the error,
      which probably isn't ideal behaviour. Oh, well.
 
+     2003/05/20: Automatic reallocation of output buffer in listsrings() by
+                 Sebastian Schmidt <yath@yath.eu.org>.
+
 */
 
 /* ========== */
@@ -60,7 +63,6 @@
 #include <stdarg.h>
 #include <stdlib.h>
 #include <string.h>
-#include <time.h>
 
 #ifdef UNIX
 #include <sys/types.h>
@@ -74,11 +76,10 @@ int skptok(int j, char *ptr);
 int evcmp(const void *e1, const void *e2);
 char *deffname(void);
 
-struct event *dir_include(char *dir, char *parm);
+struct event *dir_include(char *dir, char *parm, const struct date *today);
 
 /* ========== Global variables */
 
-struct date today;
 int iDWarn   = DEF_WARN;
 int iMaxWarn = MAX_WARN;
 int iMinWarn = MIN_WARN;
@@ -111,7 +112,7 @@ const char *NUMBERS[10] = {"zero","one","two","three","four","five","six","seven
 struct {
   const char *text;
   int len;
-  struct event *(*func)(char*,char*);
+  struct event *(*func)(char*,char*,const struct date*);
 } directlist[] = {
   { "include", 0, dir_include },
   { NULL, 0, NULL }
@@ -122,30 +123,42 @@ struct {
 /* compare the first strlen(a) characters of a and b */
 #define strbegcmp(a,b) strncmp(a,b,strlen(a))
 
+/* like strcat(), but lets the buffer automagically grow :-)
+ * (needs local variable "size" with the buffer size) */
+#define append(where, what) do {				    \
+    if (strlen(what) > (size - strlen(where) - 1)) {		    \
+      where = xrealloc(where, size + 128 + strlen(what));	    \
+      size += 128 + strlen(what);				    \
+    }								    \
+    strcat(where, what);					    \
+  } while(0)
+
 /* ========== */
 
-/* outputs delta(d) in days, weeks, months, etc to the string s */
-int tdelta(char *buf, struct date *d)
-{
+/* returns delta(d) in days, weeks, months, etc
+ * the returned buffer is malloc()ed, do not forget to free() it */
+char *tdelta(struct date *d, const struct date *today) {
   int dy, wk, mn, yr;
-  char *ptr;
+  char *tmp;
+  char *buf = xmalloc(128);
+  int size = 128;
+  *buf = 0;
   
-  ptr = buf;
-  switch (delta(d)) {
+  switch (delta(d, today)) {
   case 0:
-    sprintf(buf, "today");
-    return 5;
+    append(buf, "today");
+    return buf;
   case 1:
-    sprintf(buf, "tomorrow");
-    return 8;
+    append(buf, "tomorrow");
+    return buf;
   default:
     /* like delta(), we ignore the year */
-    yr=-before(*d,today);
-    mn=d->month - today.month;
-    dy=d->day - today.day;
+    yr=-before(*d,*today);
+    mn=d->month - today->month;
+    dy=d->day - today->day;
     
     if (dy < 0) {
-      dy += mlen(today.month, today.year);
+      dy += mlen(today->month, today->year);
       mn--;
     }
     if (mn < 0) {
@@ -156,107 +169,139 @@ int tdelta(char *buf, struct date *d)
     wk = (dy/7);
     dy%=7;
     
-    ptr += sprintf(ptr, "in ");
-    ptr += ttime(ptr, yr, mn, wk, dy);
-    if (*(ptr-1) == 's')
-      ptr += sprintf(ptr, "'");
+    append(buf, "in ");
+    tmp = ttime(yr, mn, wk, dy);
+    append(buf, tmp);
+    free(tmp);
+    
+    if (*(buf + strlen(buf) - 1) == 's')
+      append(buf, "'");
     else
-      ptr += sprintf(ptr, "'s");
-    ptr += sprintf(ptr, " time");
+      append(buf, "'s");
 
-    return ptr-buf;
+    append(buf, " time");
+
+    return buf;
   }
 }
 
-
 #ifdef NUMS_AS_WORDS
-#define donum(n,txt) \
-    if (n > 0) { \
-      if (n < 10) \
-	ptr+=sprintf(ptr, "%s " txt "%s", NUMBERS[n], (n == 1 ? "" : "s")); \
-      else \
-        ptr+=sprintf(ptr, "%d " txt "s", n); \
-      terms--; \
-      if (orgterms > 1) { \
-	if (terms == 1) ptr += sprintf(ptr, " and "); \
-	else if (terms > 1) ptr += sprintf(ptr, ", "); \
-      } \
-    }
+#define donum(n,txt) do {                          \
+      if (n > 0) {                                 \
+         if (n < 10) {                             \
+            append(buf, NUMBERS[n]);               \
+            append(buf, " " txt);                  \
+               if (n != 1) {                       \
+                  append(buf, "s");                \
+               }                                   \
+         } else {                                  \
+            snprintf(tmp, sizeof(tmp), "%d", n);	\
+            append(buf, tmp);                      \
+            append(buf, " " txt "s");              \
+         }                                         \
+         terms--;                                  \
+         if (orgterms > 1) {                       \
+	         if (terms == 1)                        \
+               append(buf, " and ");               \
+            else if (terms > 1)                    \
+               append(buf, ", ");                  \
+         }                                         \
+      }                                            \
+   } while(0)
 #else /* NUMS_AS_WORDS */
-#define donum(n,txt) \
-    if (n > 0) { \
-      ptr+=sprintf(ptr, "%d " txt "%s", n, (n == 1 ? "" : "s")); \
-      terms--; \
-      if (orgterms > 1) { \
-	if (terms == 1) ptr += sprintf(ptr, " and "); \
-	else if (terms > 1) ptr += sprintf(ptr, ", "); \
-      } \
-    }
+#define donum(n,txt) do {                       \
+      if (n > 0) {                              \
+         snprintf(tmp, sizeof(tmp), "%d", n);   \
+         append(buf, tmp);                      \
+         append(buf, " " txt);                  \
+         if (n != 1)                            \
+            append(buf, "s");                   \
+         terms--;                               \
+         if (orgterms > 1) {                    \
+            if (terms == 1)                     \
+               append(buf, " and ");            \
+            else if (terms > 1)                 \
+               append(buf, ", ");               \
+         }                                      \
+      }                                         \
+   } while(0)
 #endif /* NUMS_AS_WORDS */
 
-int ttime(char *buf, int yr, int mn, int wk, int dy)
-{
-  char *ptr;
+/* returns allocated buffer, don't forget to free() */
+char *ttime(int yr, int mn, int wk, int dy) {
+  char *buf = xmalloc(128);
+  int size = 128;
   int terms, orgterms;
+  char tmp[128];
   
-  ptr=buf;
-  terms=orgterms=(yr!=0)+(mn!=0)+(wk!=0)+(dy!=0);
+  *buf = 0; /* Initialize buffer */
+  terms = orgterms = (yr!=0) + (mn!=0) + (wk!=0) + (dy!=0);
 
   donum(yr, "year");
   donum(mn, "month");
   donum(wk, "week");
   donum(dy, "day");
 
-  return ptr-buf;
+  return buf;
 }
 #undef donum
-    
+
 /* lists the birthdays in their string format, one by one, and passes the string
    to a function. */
-void liststrings(struct event *evl, prnfunc outf)
-{
-  int i,j;
-  char buf[128];
-  char *ptr;
-  
-  for (i = 0; evl[i].text != NULL; i++) {
-    *buf=0;
-    ptr=buf;
-    
-    if (evl[i].warn == -1 && delta(&(evl[i].date))==0) {
-      ptr+=sprintf(ptr, "%s", evl[i].text);
-    } else if (evl[i].enddate.day == 0) {
-      if (delta(&(evl[i].date)) <= warnperiod(evl[i])) {
-	ptr += sprintf(ptr, "%s ", evl[i].text);
-	ptr += tdelta(ptr, &(evl[i].date));
+void liststrings(struct event *evl, prnfunc outf, const struct date *today) {
+   int i,j;
+   char *buf, *tmp;
+   int size;
+   
+   for (i = 0; evl[i].text != NULL; i++) {
+      buf = xmalloc(128);
+      *buf = '\0';
+      size = 128;
+      
+      if (evl[i].warn == -1 && delta(&(evl[i].date), today)==0) {
+         append(buf, evl[i].text);
+      } else if (evl[i].enddate.day == 0) {
+	if (delta(&(evl[i].date), today) <= warnperiod(evl[i])) {
+            append(buf, evl[i].text);
+            append(buf, " ");
+            tmp = tdelta(&(evl[i].date), today);
+            append(buf, tmp);
+            free(tmp);
+         }
+      } else {
+	if (delta(&(evl[i].date), today) <= warnperiod(evl[i])) {
+            append(buf, evl[i].text);
+            append(buf, " for ");
+	         /* +1 because, if the difference between two dates is one day,
+	            then the length of an event on those days is two days */
+	    j = ddiff(&(evl[i].date),&(evl[i].enddate), today) + 1;
+	         tmp = ttime(0, 0, j/7, j%7);
+            append(buf, tmp);
+            free(tmp);
+            append(buf, " ");
+            tmp = tdelta(&(evl[i].date), today);
+            append(buf, tmp);
+	} else if (delta(&(evl[i].enddate), today) <= warnperiod(evl[i])) {
+            append(buf, evl[i].text);
+            append(buf, " ");
+	    j = delta(&(evl[i].enddate), today);
+	         if (j) {
+               append(buf, "for ");
+	            tmp = ttime(0, 0, j/7, j%7);
+               append(buf, tmp);
+               free(tmp);
+               append(buf, " longer");
+	         } else {
+               append(buf, "finishes today");
+	         }
+         }
       }
-    } else {
-      if (delta(&(evl[i].date)) <= warnperiod(evl[i])) {
-	ptr += sprintf(ptr, "%s for ", evl[i].text);
-	/* +1 because, if the difference between two dates is one day,
-	   then the length of an event on those days is two days */
-	j = ddiff(&(evl[i].date),&(evl[i].enddate)) + 1;
-	ptr += ttime(ptr, 0, 0, j/7, j%7);
-	*(ptr++) = ' ';
-	*ptr=0;
-	ptr += tdelta(ptr, &(evl[i].date));
-      } else if (delta(&(evl[i].enddate)) <= warnperiod(evl[i])) {
-	ptr += sprintf(ptr, "%s ", evl[i].text);
-	j = delta(&(evl[i].enddate));
-	if (j) {
-	  ptr += sprintf(ptr, "for ");
-	  ptr += ttime(ptr, 0, 0, j/7, j%7);
-	  ptr += sprintf(ptr, " longer");
-	} else {
-	  ptr += sprintf(ptr, "finishes today");
-	}
+      if (*buf) {
+         append(buf, ".");
+         outf(buf);
       }
-    }
-    if (*buf) {
-      sprintf(buf + strlen(buf), ".");
-      outf(buf);
-    }
-  }
+      free(buf);
+   }
 }
 
 char *
@@ -287,13 +332,8 @@ evcmp(const void *p1, const void *p2)
      have passed the start date but not yet the end date, and so we should
      display the enddate; otherwise, we should display the start date */
   
-  d1=delta(&(e1->date));
-  if (e1->enddate.day && delta(&(e1->enddate)) < d1)
-    d1=delta(&(e1->enddate));
-
-  d2=delta(&(e2->date));
-  if (e2->enddate.day && delta(&(e2->enddate)) < d2)
-    d2=delta(&(e2->enddate));
+  d1 = e1->fromtoday;
+  d2 = e2->fromtoday;
 
   if (d1 < d2) return -1;
   if (d1 > d2) return  1;
@@ -304,11 +344,11 @@ evcmp(const void *p1, const void *p2)
 /* difference in days between two dates */
 /* it is assumed that D1 < D2, and so the result is always positive */
 
-unsigned ddiff(struct date *D1, struct date *D2)
+unsigned ddiff(struct date *D1, struct date *D2, const struct date *today)
 {
   struct date d1,d2;
   int dd,m;
-  
+
   /* make working copies */
   d1=*D1;
   d2=*D2;
@@ -328,10 +368,12 @@ unsigned ddiff(struct date *D1, struct date *D2)
 	  d2.year=d1.year+1;
       }
     } else { /* both years zero */
+      /* NB we do actually have to use today's year, to get leap years
+	 right. */
       if (before(d1,d2))
-	d1.year=d2.year=today.year;
+	d1.year=d2.year=today->year;
       else {
-	d1.year=today.year;
+	d1.year=today->year;
 	d2.year=d1.year+1;
       }
     }
@@ -354,43 +396,43 @@ unsigned ddiff(struct date *D1, struct date *D2)
 }
 
 /* actually until the next anniversary of ... */
-unsigned delta(struct date *date)
+unsigned delta(struct date *date, const struct date *today)
 {
   struct date d;
   unsigned dt, mn;
   
   memcpy(&d, date, sizeof(struct date));
   
-  if (before(d, today)) /* past the end of the year */
+  if (before(d, *today)) /* past the end of the year */
     {
       d.year = 1;
     } else {
       d.year = 0;
     }
     
-  for (mn = today.month, dt=0; mn < d.month + 12*d.year; mn++)
-    dt += mlen(((mn-1)%12) + 1,today.year + mn/12);
+  for (mn = today->month, dt=0; mn < d.month + 12*d.year; mn++)
+    dt += mlen(((mn-1)%12) + 1,today->year + mn/12);
   
-  dt -= today.day;
+  dt -= today->day;
   dt += d.day;
   
   return dt;  
 }
 
-void gettoday(void)
-{
-  struct tm *tm;
-  time_t t;
-
-  time(&t);
-  tm = localtime(&t);
-  today.day = tm->tm_mday;
-  today.month = tm->tm_mon + 1; /* 1-12 instead of 0-11 */
-  today.year = tm->tm_year;
-  today.year += 1900;
+void scan_date( const char *str, struct date *date) {
+  int n;
+  n = sscanf(str, "%d/%d/%d", &(date->day), &(date->month), &(date->year));
+  if ( n < 3 ) {
+    /* didn't read the year */
+    date->year=0;
+  }
+  else {
+    /* if our year is only two digits, add 1900 to it ... */
+    if(date->year < 100) date->year+=1900;
+  }
 }
-  
-struct event *readlist(char *fname)
+
+struct event *readlist(char *fname, const struct date *today)
 {
   FILE *file;
   int i,j,k,l,d;
@@ -399,14 +441,11 @@ struct event *readlist(char *fname)
   char *ptr;
   unsigned flags;
   struct event *nevl;
-  
   /* initialise */
   if (fname==NULL) {
     fname=deffname();
   }
   
-  gettoday();
-
   if (fname[0] == '-' && fname[1] == 0) {
     /* read from stdin */
     file=stdin;
@@ -443,7 +482,7 @@ struct event *readlist(char *fname)
 		*/
 	if (directlist[j].len == ptr-(buf+1)-1 && !strncmp(directlist[j].text,(buf+1),directlist[j].len))
 	  {
-	    nevl = directlist[j].func((buf+1),ptr);
+	    nevl = directlist[j].func((buf+1),ptr,today);
 	    k=1;
 	    break;
 	  }
@@ -475,12 +514,8 @@ struct event *readlist(char *fname)
       
     *(ptr++) = 0;  
 
-    j = sscanf(ptr, "%d/%d/%d", &(evl[i].date.day), &(evl[i].date.month), &(evl[i].date.year));
-    /* if our year is only two digits, add 1900 to it ... */
-    if(evl[i].date.year < 100) evl[i].date.year+=1900;
-    /* ... unless it wasn't read, in which case set it to zero */
-    if(j==2) evl[i].date.year=0;
-      
+    scan_date( ptr, &(evl[i].date) );
+
     /* parse flags */
 
     evl[i].warn=iDWarn;
@@ -529,22 +564,27 @@ struct event *readlist(char *fname)
     case F_TBIRTHDAY:
     default: /* assume it's a birthday */
       if (evl[i].date.year != 0) {
-	sprintf(buf2, "%s is %d years old", buf, ydelta(evl[i].date, today));
+	int tmp_age=ydelta(evl[i].date, *today);
+	if (tmp_age!=1) {
+	  sprintf(buf2, "%s is %d years old", buf, tmp_age);
+	} else {
+	  sprintf(buf2, "%s is %d year old", buf, tmp_age);
+	}
       } else {
 	sprintf(buf2, "%s has a birthday", buf);
       }
       break;
     case F_TANNIVERSARY:
       if (evl[i].date.year != 0) {
-	sprintf(buf2, "%s %d years ago", buf, ydelta(evl[i].date, today));
+	sprintf(buf2, "%s %d years ago", buf, ydelta(evl[i].date, *today));
       } else {
 	strcpy(buf2, buf);
       }
       break;
     case F_TEVENT:
       /* if a year was specified, and this warning isn't for it, ignore! */
-      if ((evl[i].date.year != 0 && ydelta(evl[i].date, today) != 0) &&
-	  (evl[i].enddate.year == 0 || ydelta(evl[i].enddate, today) != 0)) {
+      if ((evl[i].date.year != 0 && ydelta(evl[i].date, *today) != 0) &&
+	  (evl[i].enddate.year == 0 || ydelta(evl[i].enddate, *today) != 0)) {
 	i--;
 	continue;
       }
@@ -552,8 +592,8 @@ struct event *readlist(char *fname)
       break;
     case F_TMESSAGE:
       /* Like an event, except that it only comes up on the given date, and no text at all is appended */
-      if ((evl[i].date.year != 0 && ydelta(evl[i].date, today) != 0) &&
-	  (evl[i].enddate.year == 0 || ydelta(evl[i].enddate, today) != 0)) {
+      if ((evl[i].date.year != 0 && ydelta(evl[i].date, *today) != 0) &&
+	  (evl[i].enddate.year == 0 || ydelta(evl[i].enddate, *today) != 0)) {
 	i--;
 	continue;
       }
@@ -562,6 +602,16 @@ struct event *readlist(char *fname)
       break;
     }      
     evl[i].text = strdup(buf2);
+
+    /* precalculate the "priority", the time until we next need to
+       display something about this event. */
+    {
+      unsigned d;
+      d=delta(&(evl[i].date), today);
+      if (evl[i].enddate.day && delta(&(evl[i].enddate), today) < d)
+	d=delta(&(evl[i].enddate), today);
+      evl[i].fromtoday = d;
+    }
   }
 
   evl = (struct event *) xrealloc(evl, sizeof(struct event) * (i + 1));
@@ -573,7 +623,6 @@ struct event *readlist(char *fname)
 
   /* NB uses i from above */
   qsort(evl, i, sizeof(struct event), evcmp);
-  
   return evl;
 }
 
@@ -588,11 +637,11 @@ skptok(int j, char *ptr)
 
 /* ---------------------------------------------------------------------- */
 /* Functions to parse input file directives */
-struct event *dir_include(char *dir, char *parm)
+struct event *dir_include(char *dir, char *parm, const struct date *today)
 {
   struct event *evl;
 
-  evl=readlist(strdup(parm));
+  evl=readlist(strdup(parm), today);
 
   return evl;
 }
