@@ -3,14 +3,14 @@
 
 	Note: the x* functions are lifted straight from the GNU libc info docs
 	
-	$Id: xmalloc.c,v 1.2 1999/01/16 17:08:59 andy Exp $
+	$Id: xmalloc.c,v 1.2 2005/07/19 18:50:00 andymort Exp $
 */
 
 #include <stdlib.h>
 #include <stdio.h>
 #include <stdarg.h>
 
-#include "junk.h"
+#include "xmalloc.h"
 
 void *xmalloc (size_t size)
 {
